# -*- coding: utf-8 -*-
#
#   This file is part of the rnasamba package, available at:
#   https://github.com/apcamargo/RNAsamba
#
#   Rnasamba is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#   Contact: antoniop.camargo@gmail.com

from keras.preprocessing.sequence import pad_sequences

from rnasamba.core import sequences


class RNAsambaInput:
    def __init__(self, fasta_file, maxlen=3000):
        self._tokenized_sequences = sequences.read_fasta(fasta_file, tokenize=True)
        self._nucleotide_sequences = sequences.read_fasta(fasta_file, tokenize=False)
        self._aa_dict = {
            'A': 4, 'C': 18, 'D': 12, 'E': 3, 'F': 14, 'G': 5, 'H': 16, 'I': 13, 'K': 9, 'L': 1,
            'M': 19, 'N': 15, 'P': 6, 'Q': 11, 'R': 8, 'S': 2, 'T': 10, 'V': 7, 'W': 20, 'X': 21,
            'Y': 17
        }
        self._orfs = self.get_orfs()
        self.protein_seqs = [orf[2] for orf in self._orfs]
        self.maxlen = maxlen
        self.protein_maxlen = int(maxlen/3)
        self.nucleotide_input = self.get_nucleotide_input()
        self.kmer_frequency_input = self.get_kmer_frequency_input()
        self.orf_indicator_input = self.get_orf_indicator_input()
        self.protein_input = self.get_protein_input()
        self.aa_frequency_input = self.get_aa_frequency_input()
        self.sequence_name = [seq[1] for seq in self._nucleotide_sequences]

    def get_orfs(self):
        orfs = [sequences.longest_orf(seq[0]) for seq in self._nucleotide_sequences]
        return orfs

    def get_nucleotide_input(self):
        nucleotide_input = [i[0] for i in self._tokenized_sequences]
        nucleotide_input = pad_sequences(nucleotide_input, padding='post', maxlen=self.maxlen)
        return nucleotide_input

    def get_kmer_frequency_input(self):
        kmer_frequency_input = sequences.kmer_frequency(self._nucleotide_sequences)
        return kmer_frequency_input

    def get_orf_indicator_input(self):
        orf_indicator_input = sequences.orf_indicator(self._orfs, self.maxlen)
        return orf_indicator_input

    def get_protein_input(self):
        protein_input = []
        for protein_seq in self.protein_seqs:
            protein_numeric = [self._aa_dict[aa] for aa in protein_seq]
            protein_input.append(protein_numeric)
        protein_input = pad_sequences(protein_input, padding='post', maxlen=self.protein_maxlen)
        return protein_input

    def get_aa_frequency_input(self):
        aa_frequency_input = sequences.aa_frequency(self._aa_dict, self._orfs)
        return aa_frequency_input



# -*- coding: utf-8 -*-
#
#   This file is part of the rnasamba package, available at:
#   https://github.com/apcamargo/RNAsamba
#
#   Rnasamba is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#   Contact: antoniop.camargo@gmail.com

import itertools
import re
from collections import Counter

import numpy as np
from Bio import Seq, SeqIO
from keras.preprocessing.sequence import pad_sequences
from keras.utils import to_categorical


def read_fasta(filename, tokenize=False):
    seqs = []
    with open(filename) as handle:
        if tokenize:
            for record in SeqIO.parse(handle, 'fasta'):
                sequence_str = str(record.seq).upper().replace('U', 'T')
                sequence_name = record.description
                seqs.append((tokenize_dna(sequence_str), sequence_name))
        else:
            for record in SeqIO.parse(handle, 'fasta'):
                sequence_str = str(record.seq).upper().replace('U', 'T')
                sequence_name = record.description
                seqs.append((sequence_str, sequence_name))
    return seqs


def tokenize_dna(seq):
    lookup = dict(zip('NATCG', range(5)))
    if not seq:
        token = [0]
    else:
        token = [lookup[c] for c in seq if c in lookup]
    return token


def longest_orf(input_seq):
    start_codon = re.compile('ATG')
    longest = (0, 0, '')
    for m in start_codon.finditer(input_seq):
        putative_orf = input_seq[m.start():]
        # Add trailing Ns to make the sequence length a multiple of three:
        putative_orf = putative_orf + 'N'*(3-len(putative_orf) % 3)
        protein = Seq.Seq(putative_orf).translate(to_stop=True)
        if len(protein) > longest[0]:
            longest = (len(protein),
                       m.start(),
                       str(protein))
    return longest


def orf_indicator(orfs, maxlen):
    orf_indicator = []
    for orf in orfs:
        orf_vector = np.zeros(maxlen)
        if orf[0] > 1:
            orf_vector[orf[1]:orf[1]+orf[0]*3] = 1
        orf_indicator.append(orf_vector)
    orf_indicator = pad_sequences(orf_indicator, padding='post', maxlen=maxlen)
    orf_indicator = to_categorical(orf_indicator, num_classes=2)
    orf_indicator = np.stack(orf_indicator)
    return orf_indicator


def count_kmers(read, k):
    counts = {}
    num_kmers = len(read) - k + 1
    for i in range(num_kmers):
        kmer = read[i:i+k]
        if kmer not in counts:
            counts[kmer] = 0
        counts[kmer] += 1
    return counts


def kmer_frequency(nucleotide_sequences):
    kmer_frequency = []
    bases = ['A', 'T', 'C', 'G']
    kmer_lengths = [2, 3, 4]
    for nucleotide_seq in nucleotide_sequences:
        matches = [bases, bases]
        sequence_kmer_frequency = []
        for current_length in kmer_lengths:
            current_seq = nucleotide_seq[0]
            total_kmers = len(current_seq)-(current_length-1)
            kmer_count = count_kmers(current_seq, current_length)
            for match in itertools.product(*matches):
                current_kmer = ''.join(match)
                if current_kmer in kmer_count:
                    sequence_kmer_frequency.append(kmer_count[current_kmer]/(total_kmers))
                else:
                    sequence_kmer_frequency.append(0)
            matches.append(bases)
        kmer_frequency.append(sequence_kmer_frequency)
    kmer_frequency = np.stack(kmer_frequency)
    return kmer_frequency


def aa_frequency(aa_dict, orfs):
    aa_numeric = list(aa_dict.values())
    aa_numeric.sort()
    aa_frequency = []
    for orf in orfs:
        protein_seq = orf[2]
        protein_numeric = [aa_dict[aa] for aa in protein_seq]
        aa_count = Counter(protein_numeric)
        protein_len = max(len(protein_numeric), 1)
        freq = []
        for aa in aa_numeric:
            if aa in aa_count:
                freq.append(aa_count[aa]/protein_len)
            else:
                freq.append(0.)
        aa_frequency.append(freq)
    aa_frequency = np.stack(aa_frequency)
    return aa_frequency



# -*- coding: utf-8 -*-
#
#   This file is part of the rnasamba package, available at:
#   https://github.com/apcamargo/RNAsamba
#
#   Rnasamba is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#   Contact: antoniop.camargo@gmail.com

import argparse
import sys

from rnasamba import RNAsambaClassificationModel, RNAsambaTrainModel

def classify(output_file, fasta_file, weights, protein_fasta, verbose):
    """Classify sequences from a input FASTA file."""
    classification = RNAsambaClassificationModel(fasta_file, weights, verbose=verbose)
    classification.write_classification_output(output_file)
    if protein_fasta:
        classification.output_protein_fasta(protein_fasta)

def train(output_file, coding_file, noncoding_file, early_stopping, batch_size, epochs, verbose):
    """Train a classification model from training data and saves the weights into a HDF5 file."""
    trained = RNAsambaTrainModel(coding_file, noncoding_file, early_stopping=early_stopping,
                                 batch_size=batch_size, epochs=epochs, verbose=verbose)
    trained.model.save_weights(output_file)

def classify_cli():
    parser = argparse.ArgumentParser(description='Classify sequences from a input FASTA file.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('output_file',
                        help='output TSV file containing the results of the classification.')
    parser.add_argument('fasta_file',
                        help='input FASTA file containing transcript sequences.')
    parser.add_argument('weights',
                        nargs='+', help='input HDF5 file(s) containing weights of a trained RNAsamba network (if more than a file is provided, an ensembling of the models will be performed).')
    parser.add_argument('-p', '--protein_fasta',
                        help='output FASTA file containing translated sequences for the predicted coding ORFs.')
    parser.add_argument('-v', '--verbose',
                        default=0, type=int, choices=[0, 1],
                        help='print the progress of the classification. 0 = silent, 1 = current step.')
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)
    args = parser.parse_args()
    classify(**vars(args))

def train_cli():
    parser = argparse.ArgumentParser(description='Train a new classification model.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('output_file',
                        help='output HDF5 file containing weights of the newly trained RNAsamba network.')
    parser.add_argument('coding_file',
                        help='input FASTA file containing sequences of protein-coding transcripts.')
    parser.add_argument('noncoding_file',
                        help='input FASTA file containing sequences of noncoding transcripts.')
    parser.add_argument('-s', '--early_stopping',
                        default=0, type=int, help='number of epochs after lowest validation loss before stopping training (a fraction of 0.1 of the training set is set apart for validation and the model with the lowest validation loss will be saved).')
    parser.add_argument('-b', '--batch_size',
                        default=128, type=int, help='number of samples per gradient update.')
    parser.add_argument('-e', '--epochs',
                        default=40, type=int, help='number of epochs to train the model.')
    parser.add_argument('-v', '--verbose',
                        default=0, type=int, choices=[0, 1, 2, 3],
                        help='print the progress of the training. 0 = silent, 1 = current step, 2 = progress bar, 3 = one line per epoch.')
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)
    args = parser.parse_args()
    train(**vars(args))



# -*- coding: utf-8 -*-
#
#   This file is part of the rnasamba package, available at:
#   https://github.com/apcamargo/RNAsamba
#
#   Rnasamba is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#   Contact: antoniop.camargo@gmail.com

import numpy as np
import tensorflow as tf
from keras.engine.topology import Layer
from keras.layers import Concatenate, LeakyReLU, SpatialDropout1D
from keras.layers.convolutional import Conv1D
from keras.layers.core import Dropout
from keras.layers.normalization import BatchNormalization
from keras.layers.pooling import MaxPooling1D
from keras.regularizers import l2


class RNAsambaAttention(Layer):
    def __init__(self, **kwargs):
        self.supports_masking = True
        super(RNAsambaAttention, self).__init__(**kwargs)

    def compute_mask(self, input, input_mask=None):
        if input_mask is not None:
            return input_mask
        else:
            return None

    def build(self, input_shape):
        self.batch_size = input_shape[0][0]
        self.num_channels_input = input_shape[0][1]
        self.fulloutput = input_shape[2][1]
        self.w_1 = self.add_weight(shape=(self.num_channels_input, 2),
                                   initializer='glorot_normal',
                                   trainable=True,
                                   name='w_1')
        self.w_2 = self.add_weight(shape=(1, 2),
                                   initializer='glorot_normal',
                                   trainable=True,
                                   name='w_2')
        super(RNAsambaAttention, self).build(input_shape)

    def call(self, inputs, mask=None):
        orf_length_branch = inputs[0]
        first_branch = inputs[1]
        second_branch = inputs[2]
        alpha = tf.matmul(orf_length_branch, self.w_1) + self.w_2
        alpha = tf.nn.softmax(alpha)
        weighted_first_branch = tf.multiply(tf.expand_dims(alpha[:, 0], axis=-1), first_branch)
        weighted_second_branch = tf.multiply(tf.expand_dims(alpha[:, 1], axis=-1), second_branch)
        attention_output = weighted_first_branch + weighted_second_branch
        return attention_output

    def compute_output_shape(self, input_shape):
        return input_shape[0][0], self.fulloutput


def IGLOO1D(input_layer, nb_patches, nb_filters_conv1d, patch_size=4,
            padding_style='causal', add_batchnorm=False, stretch_factor=1,
            nb_stacks=1, l2reg=0.00001, conv1d_kernel=3, max_pooling_kernel=1, DR=0.0,
            add_residual=True, nb_sequences=-1, build_backbone=False, psy=0.15, dilf_factor=3):
    M = IGLOO(
        input_layer, nb_patches, nb_filters_conv1d, patch_size=patch_size,
        padding_style=padding_style, add_batchnorm=add_batchnorm, nb_stacks=nb_stacks,
        l2reg=l2reg, conv1d_kernel=conv1d_kernel, max_pooling_kernel=max_pooling_kernel,
        DR=DR, build_backbone=build_backbone, dilf_factor=dilf_factor)
    return M


def IGLOO(input_layer, nb_patches, nb_filters_conv1d, patch_size=4,
          padding_style='causal', add_batchnorm=False, nb_stacks=1, l2reg=0.00001, conv1d_kernel=3,
          max_pooling_kernel=1, DR=0.0, build_backbone=True, dilf_factor=3):
    LAYERS = []
    x = Conv1D(nb_filters_conv1d, conv1d_kernel, padding=padding_style)(input_layer)
    if add_batchnorm:
        x = BatchNormalization(axis=-1)(x)
    x = LeakyReLU(alpha=0.1)(x)
    x = SpatialDropout1D(DR)(x)
    x = MaxPooling1D(pool_size=max_pooling_kernel, strides=None, padding='valid')(x)
    LAYERS.append(PatchyLayerCNNTopLast(patch_size, nb_patches,
                                        DR, build_backbone=build_backbone, l2reg=l2reg)(x))
    if nb_stacks > 1:
        for extra_l in range(nb_stacks-1):
            x = Conv1D(nb_filters_conv1d, conv1d_kernel,
                       padding=padding_style, dilation_rate=dilf_factor)(x)
            if add_batchnorm:
                x = BatchNormalization(axis=-1)(x)
            x = LeakyReLU(alpha=0.1)(x)
            x = SpatialDropout1D(DR)(x)
            LAYERS.append(PatchyLayerCNNTopLast(patch_size, nb_patches,
                                                DR, build_backbone=build_backbone, l2reg=l2reg)(x))
    if nb_stacks > 1:
        MPI = Concatenate()(LAYERS)
    else:
        MPI = LAYERS[0]
    return MPI


class PatchyLayerCNNTopLast(Layer):
    def __init__(
            self, patch_size, nb_patches, DR, initializer='glorot_normal', build_backbone=True,
            l2reg=0.000001, activation='relu', **kwargs):
        self.supports_masking = True
        self.nb_patches = nb_patches
        self.patch_size = patch_size
        self.DR = DR
        self.initializer = initializer
        self.kernel_causal = 4
        self.activation = activation
        self.l2reg = l2reg
        self.outsize = 100
        self.build_backbone = build_backbone
        super(PatchyLayerCNNTopLast, self).__init__(**kwargs)

    def compute_mask(self, input, input_mask=None):
        if input_mask is not None:
            return input_mask
        else:
            return None

    def PatchyLayerCNNTopLast_initializer(self, shape, dtype=None):
        M = gen_filters_igloo_newstyle1Donly(
            self.patch_size, self.nb_patches, self.vector_size, self.num_channels_input,
            build_backbone=self.build_backbone)
        M.astype(int)
        return M

    def build(self, input_shape):
        self.batch_size = input_shape[0]
        self.vector_size = input_shape[1]
        self.num_channels_input = input_shape[2]
        self.mshapy = (int(self.nb_patches), int(self.patch_size*self.num_channels_input), 2)
        self.patches = self.add_weight(shape=(int(self.nb_patches), self.patch_size, 1),
                                       initializer=self.PatchyLayerCNNTopLast_initializer,
                                       trainable=False,
                                       name='random_patches', dtype=np.int32)
        self.W_MULT = self.add_weight(
            shape=(1, self.nb_patches, self.patch_size, self.num_channels_input),
            initializer=self.initializer, trainable=True, regularizer=l2(self.l2reg),
            name='W_MULT')
        self.W_BIAS = self.add_weight(shape=(1, int(self.nb_patches/1)),
                                      initializer=self.initializer,
                                      trainable=True,
                                      regularizer=l2(self.l2reg),
                                      name='W_BIAS')
        super(PatchyLayerCNNTopLast, self).build(input_shape)

    def call(self, y, mask=None):
        y = tf.transpose(y, [1, 2, 0])
        M = tf.gather_nd(y, self.patches)
        PATCH_tensor1 = tf.transpose(M, [3, 0, 1, 2])
        MPI = tf.multiply(self.W_MULT, PATCH_tensor1)
        MPI = tf.reduce_sum(MPI, axis=[2, 3])
        MPI = tf.add(MPI, self.W_BIAS)
        MPI = LeakyReLU(alpha=0.1)(MPI)
        MPI = Dropout(self.DR)(MPI)
        return MPI

    def compute_output_shape(self, input_shape):
        return input_shape[0], self.nb_patches


def gen_filters_igloo_newstyle1Donly(
        patch_size, nb_patches, vector_size, num_channels_input_reduced,
        nb_stacks=1, build_backbone=True, consecutive=False, nb_sequences=-1):
    OUTA = []
    vector_size = int(vector_size)
    for step in range(vector_size):
        if (step != vector_size-1):
            continue
        COLLECTA = []
        if step < patch_size:
            for kk in range(nb_patches):
                randy_H = np.random.choice(range(step+1), patch_size, replace=True)
                first = []
                for pp in randy_H:
                    first.append([pp])
                COLLECTA.append(first)
            OUTA.append(COLLECTA)
        else:
            if build_backbone:
                maximum_its = int((step/(patch_size-1))+1)
                if maximum_its > nb_patches:
                    print('nb_patches too small, recommende above:', maximum_its)
                    sys.exit()
                for jj in range(maximum_its):
                    if iter == 0:
                        randy_H = [step-pp for pp in range(patch_size)]
                    else:
                        randy_H = [max(step - (jj * (patch_size - 1)) - pp, 0)
                                   for pp in range(patch_size)]
                    first = []
                    for pp in randy_H:
                        first.append([pp])
                    COLLECTA.append(first)
                rest_iters = max(nb_patches-maximum_its, 0)
                for itero in range(rest_iters):
                    if not consecutive:
                        randy_B = np.random.choice(range(step+1), patch_size, replace=False)
                    else:
                        uniq = np.random.choice(
                            range(max(0, step + 1 - patch_size + 1)),
                            1, replace=False)
                        randy_B = [uniq[0]+pp for pp in range(patch_size)]
                    first = []
                    sorting = True
                    if sorting:
                        randy_B = sorted(randy_B)
                    for pp in randy_B:
                        first.append([pp])
                    COLLECTA.append(first)
                COLLECTA = np.stack(COLLECTA)
                OUTA.append(COLLECTA)
            else:
                for itero in range(nb_patches):
                    if not consecutive:
                        randy_B = np.random.choice(range(step+1), patch_size, replace=False)
                    else:
                        uniq = np.random.choice(
                            range(max(0, step + 1 - patch_size + 1)),
                            1, replace=False)
                        randy_B = [uniq[0]+pp for pp in range(patch_size)]
                    first = []
                    sorting = True
                    if sorting:
                        randy_B = sorted(randy_B)
                    for pp in randy_B:
                        first.append([pp])
                    COLLECTA.append(first)
                COLLECTA = np.stack(COLLECTA)
                OUTA.append(COLLECTA)
    OUTA = np.stack(OUTA)
    OUTA = np.squeeze(OUTA, axis=0)
    return OUTA

    def compute_output_shape(self, input_shape):
        return input_shape



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   This file is part of the rnasamba package, available at:
#   https://github.com/apcamargo/RNAsamba
#
#   Rnasamba is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#   Contact: antoniop.camargo@gmail.com

from setuptools import setup, find_packages

setup(
    name='rnasamba',
    version='0.1.4',
    packages=find_packages(),
    license='GNU General Public License v3.0',
    description='A tool for computing the coding potential of RNA transcript sequences using deep learning.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=['biopython', 'numpy', 'keras>=2.1.0', 'tensorflow>=1.5.0,<2.0'],
    python_requires= '>=3.6',
    entry_points = {
        'console_scripts': [
            'rnasamba-classify=rnasamba.cli:classify_cli',
            'rnasamba-train=rnasamba.cli:train_cli'
            ],
    },
    url='https://github.com/apcamargo/RNAsamba/',
    keywords=['bioinformatics', 'coding potential', 'transcriptomics', 'machine learning', 'neural networks'],
    author='Antonio Pedro Camargo, Vsevolod Sourkov',
    author_email='antoniop.camargo@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Topic :: Software Development :: Libraries',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3',
    ],
)


# -*- coding: utf-8 -*-
#
#   This file is part of the rnasamba package, available at:
#   https://github.com/apcamargo/RNAsamba
#
#   Rnasamba is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program. If not, see <https://www.gnu.org/licenses/>.
#
#   Contact: antoniop.camargo@gmail.com

import logging

import numpy as np
import tensorflow as tf
from keras import optimizers
from keras.callbacks import EarlyStopping
from keras.layers import (Activation, Concatenate, Dense, Dropout, Embedding,
                          Input, Lambda)
from keras.layers.normalization import BatchNormalization
from keras.models import Model

from rnasamba.core.inputs import RNAsambaInput
from rnasamba.core.miniigloo import IGLOO1D, RNAsambaAttention


class RNAsambaClassificationModel:
    def __init__(self, fasta_file, weights, verbose=0):
        if verbose > 0:
            logging.basicConfig(level=logging.INFO, format='%(message)s')
        else:
            logging.basicConfig(level=logging.WARNING, format='%(message)s')
        logger = logging.getLogger()
        logger.info('- Computing network inputs.')
        self.input = RNAsambaInput(fasta_file)
        self.maxlen = self.input.maxlen
        self.protein_maxlen = self.input.protein_maxlen
        self.sequence_name = self.input.sequence_name
        self.protein_seqs = self.input.protein_seqs
        self.input_dict = {'nucleotide_layer': self.input.nucleotide_input,
                           'orf_indicator_layer': self.input.orf_indicator_input,
                           'kmer_frequency_layer': self.input.kmer_frequency_input,
                           'protein_layer': self.input.protein_input,
                           'aa_frequency_layer': self.input.aa_frequency_input}
        if len(weights) == 1:
            logger.info('- Building the model.')
            model = get_rnasamba_model(self.maxlen, self.protein_maxlen)
            logger.info('- Loading network weights.')
            model.load_weights(weights[0])
            logger.info('- Classifying sequences.')
            self.predictions = model.predict(self.input_dict)
            self.coding_score = self.predictions[:, 1]
            self.classification_label = np.argmax(self.predictions, axis=1)
            self.classification_label = ['coding' if i == 1 else 'noncoding' for i in self.classification_label]
        else:
            logger.info('- Building the models.')
            n_models = len(weights)
            models = [get_rnasamba_model(self.maxlen, self.protein_maxlen) for i in range(n_models)]
            logger.info('- Loading network weights.')
            for i in range(n_models):
                models[i].load_weights(weights[i])
            logger.info('- Classifying sequences using an ensemble of {} models.'.format(n_models))
            self.predictions = np.average([models[i].predict(self.input_dict)
                                           for i in range(n_models)], axis=0)
            self.coding_score = self.predictions[:, 1]
            self.classification_label = np.argmax(self.predictions, axis=1)
            self.classification_label = ['coding' if i == 1 else 'noncoding' for i in self.classification_label]

    def write_classification_output(self, output_file):
        with open(output_file, 'w') as handle:
            handle.write('sequence_name\tcoding_score\tclassification\n')
            for i in range(len(self.classification_label)):
                handle.write(self.sequence_name[i])
                handle.write('\t')
                handle.write('{:.5f}'.format(self.coding_score[i]))
                handle.write('\t')
                handle.write(self.classification_label[i])
                handle.write('\n')

    def output_protein_fasta(self, protein_fasta):
        with open(protein_fasta, 'w') as handle:
            for i in range(len(self.classification_label)):
                if self.classification_label[i] == 'coding':
                    if self.protein_seqs[i]:
                        handle.write('>')
                        handle.write(self.sequence_name[i])
                        handle.write('\n')
                        handle.write(self.protein_seqs[i])
                        handle.write('\n')


class RNAsambaTrainModel:
    def __init__(self, coding_file, noncoding_file, early_stopping=0, batch_size=128, epochs=40, verbose=0):
        if verbose > 0:
            verbose_keras = verbose - 1
            logging.basicConfig(level=logging.INFO, format='%(message)s')
        else:
            verbose_keras = verbose
            logging.basicConfig(level=logging.WARNING, format='%(message)s')
        logger = logging.getLogger()
        logger.info('- Computing network inputs.')
        self.coding_input = RNAsambaInput(coding_file)
        self.noncoding_input = RNAsambaInput(noncoding_file)
        self.maxlen = self.coding_input.maxlen
        self.protein_maxlen = self.coding_input.protein_maxlen
        self.labels = np.repeat([[0, 1], [1, 0]], [len(self.coding_input.sequence_name), len(
            self.noncoding_input.sequence_name)], axis=0)
        self.input_dict = {'nucleotide_layer': np.concatenate(
            [self.coding_input.nucleotide_input, self.noncoding_input.nucleotide_input]),
            'orf_indicator_layer': np.concatenate(
                [self.coding_input.orf_indicator_input, self.noncoding_input.orf_indicator_input]),
            'kmer_frequency_layer': np.concatenate(
                [self.coding_input.kmer_frequency_input, self.noncoding_input.kmer_frequency_input]),
            'protein_layer': np.concatenate(
                [self.coding_input.protein_input, self.noncoding_input.protein_input]),
            'aa_frequency_layer': np.concatenate(
                [self.coding_input.aa_frequency_input, self.noncoding_input.aa_frequency_input])}
        logger.info('- Building the model.')
        self.model = get_rnasamba_model(self.maxlen, self.protein_maxlen)
        logger.info('- Training the network.')
        if early_stopping > 0:
            logger.info('- Using early stopping. 10% of the training data will be set aside for validation.')
            seed = np.random.randint(0, 50)
            np.random.seed(seed)
            np.random.shuffle(self.labels)
            for i in self.input_dict:
                np.random.seed(seed)
                np.random.shuffle(self.input_dict[i])
            early_stop_call = EarlyStopping(monitor='val_loss', patience=early_stopping,
                                            restore_best_weights=True, verbose=verbose_keras)
            self.model.fit(self.input_dict, self.labels, callbacks=[early_stop_call],
                           validation_split=0.1, shuffle=True, batch_size=batch_size, epochs=epochs,
                           verbose=verbose_keras)
        else:
            self.model.fit(self.input_dict, self.labels, shuffle=True,
                           batch_size=batch_size, epochs=epochs, verbose=verbose_keras)


def get_rnasamba_model(maxlen, protein_maxlen):
    nucleotide_layer = Input(name='nucleotide_layer', shape=(maxlen,))
    kmer_frequency_layer = Input(name='kmer_frequency_layer', shape=(336,))
    orf_indicator_layer = Input(name='orf_indicator_layer', shape=(maxlen, 2))
    protein_layer = Input(name='protein_layer', shape=(protein_maxlen,))
    aa_frequency_layer = Input(name='aa_frequency_layer', shape=(21,))

    # Nucleotide branch (first branch):
    emb_mat_nuc = np.random.normal(0, 1, (5, 4))
    for i in range(1, 5):
        vector = np.zeros(4)
        vector[i-1] = 1
        emb_mat_nuc[i] = vector
    embedding_layer_nucleotide = Embedding(input_dim=emb_mat_nuc.shape[0], output_dim=4,
                                           weights=[emb_mat_nuc], trainable=False, mask_zero=False)
    emb_nuc = embedding_layer_nucleotide(nucleotide_layer)
    nucleotide_branch_1 = IGLOO1D(
        emb_nuc, nb_patches=900, nb_filters_conv1d=6, padding_style='same',
        add_batchnorm=True, nb_stacks=6, conv1d_kernel=1, l2reg=0.01, DR=0.30, max_pooling_kernel=8)
    nucleotide_branch_2 = IGLOO1D(
        emb_nuc, nb_patches=900, nb_filters_conv1d=6, padding_style='same',
        add_batchnorm=True, nb_stacks=6, conv1d_kernel=3, l2reg=0.01, DR=0.30, max_pooling_kernel=8)
    first_branch = Concatenate(axis=-1)([nucleotide_branch_1, nucleotide_branch_2])
    first_branch = Dense(128)(first_branch)
    first_branch = BatchNormalization()(first_branch)
    first_branch = Activation('relu')(first_branch)
    first_branch = Dropout(0.30)(first_branch)

    # k-mer frequency branch:
    kmer_branch = Dense(128)(kmer_frequency_layer)
    kmer_branch = BatchNormalization()(kmer_branch)
    kmer_branch = Activation('relu')(kmer_branch)
    kmer_branch = Dropout(0.30)(kmer_branch)
    kmer_branch = Dense(64)(kmer_branch)
    kmer_branch = BatchNormalization()(kmer_branch)
    kmer_branch = Activation('relu')(kmer_branch)
    kmer_branch = Dropout(0.30)(kmer_branch)

    # ORF branch:
    orf_length_branch = Lambda(lambda d: tf.reduce_sum(
        d[:, :, 1]/maxlen, axis=-1))(orf_indicator_layer)
    orf_length_branch = Lambda(lambda d: tf.expand_dims(d, axis=-1))(orf_length_branch)
    orf_length_branch = Lambda(lambda d: tf.tile(d, [1, 64]))(orf_length_branch)

    # Protein branch:
    emb_mat_prot = np.random.random((22, 21))
    emb_mat_prot[0] = np.zeros(21)
    for i in range(1, 22):
        vector = np.zeros(21)
        vector[i-1] = 1
        emb_mat_prot[i] = vector
    embedding_layer_protein = Embedding(input_dim=emb_mat_prot.shape[0], output_dim=21,
                                        weights=[emb_mat_prot], trainable=False, mask_zero=False)
    emb_prot = embedding_layer_protein(protein_layer)
    protein_branch = IGLOO1D(
        emb_prot, nb_patches=600, nb_filters_conv1d=8, padding_style='same',
        add_batchnorm=True, nb_stacks=6, conv1d_kernel=3, l2reg=0.13, DR=0.30, max_pooling_kernel=2)
    protein_branch = Dense(64)(protein_branch)
    protein_branch = BatchNormalization()(protein_branch)
    protein_branch = Activation('relu')(protein_branch)
    protein_branch = Dropout(0.30)(protein_branch)

    # Aminoacid frequency branch:
    aa_branch = Dense(36)(aa_frequency_layer)
    aa_branch = BatchNormalization()(aa_branch)
    aa_branch = Activation('relu')(aa_branch)
    aa_branch = Dropout(0.30)(aa_branch)
    aa_branch = Dense(36)(aa_branch)
    aa_branch = BatchNormalization()(aa_branch)
    aa_branch = Activation('relu')(aa_branch)
    aa_branch = Dropout(0.30)(aa_branch)

    # Second branch:
    second_branch = Concatenate(
        axis=-1)([kmer_branch, orf_length_branch, protein_branch, aa_branch])
    second_branch = Dense(128)(second_branch)
    second_branch = BatchNormalization()(second_branch)
    second_branch = Activation('relu')(second_branch)
    second_branch = Dropout(0.30)(second_branch)

    # Attention layer:
    attention = RNAsambaAttention()([orf_length_branch, first_branch, second_branch])
    output_layer = Dense(2, activation='softmax')(attention)
    model = Model([nucleotide_layer, orf_indicator_layer, kmer_frequency_layer,
                   protein_layer, aa_frequency_layer], output_layer)
    model.compile(loss='categorical_crossentropy',
                  optimizer=optimizers.Adam(lr=0.0025, clipnorm=1., decay=0.01),
                  metrics=['accuracy'])
    return model
